unit Unit1;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls;

type
  TForm1 = class(TForm)
    LoadSaveBtn: TButton;
    Screen: TImage;
    BackBuffer: TImage;
    GoRecBtn: TButton;
    GoWaveTraceBtn: TButton;
    PrimBtn: TButton;
    KruskalBtn: TButton;
    procedure LoadSaveBtnClick(Sender: TObject);
    procedure GoRecBtnClick(Sender: TObject);
    procedure GoWaveTraceBtnClick(Sender: TObject);
    procedure PrimBtnClick(Sender: TObject);
    procedure KruskalBtnClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  Form1: TForm1;
implementation

{$R *.dfm}

{------------------------------------------------------}

type Location = record
    left_wall, up_wall : Boolean;
end;
type Maze = array of array of Location;

const CellSize = 20;                    {      }

{------------------------------------------------------}

{   }
procedure LoadMaze(var TheMaze : Maze; FileName : string);
var f             : TextFile;            {     }
    Height, Width : Integer;             {     }
    x, y          : Integer;             {    }
    lw, uw        : Integer;             {   }
begin
    AssignFile(f, FileName);             {   }
    Reset(f);

    ReadLn(f, Width, Height);                    {     }
    SetLength(TheMaze, Width + 1, Height + 1);   {    }

    for y := 0 to Height do                     {     }
        for x := 0 to Width do
        if (y = Height) or (x = Width) then     {   -  }
        begin
            TheMaze[x, y].left_wall := true;    {    }
            TheMaze[x, y].up_wall := true;
        end
        else
        begin                                         {   }
            ReadLn(f, uw, lw);                        {   }
            TheMaze[x, y].left_wall := Boolean(lw);   {   }
            TheMaze[x, y].up_wall := Boolean(uw);     {    }
        end;                                          {   Boolean }
    CloseFile(f);
end;                                            {   }

{------------------------------------------------------}

{   }
procedure SaveMaze(TheMaze : Maze; FileName : string);
var f             : TextFile;                {     }
    Height, Width : Integer;                 {    }
    x, y          : Integer;                 {    }
begin
    AssignFile(f, FileName);                 {   }
    Rewrite(f);                              {   }

    Height := High(TheMaze[0]);              {   }
    Width := High(TheMaze);                  {    }

    WriteLn(f, Width, ' ', Height);          {       }

    for y := 0 to Height - 1 do              {    }
        for x := 0 to Width - 1 do
            WriteLn(f, Integer(TheMaze[x, y].up_wall), ' ', Integer(TheMaze[x, y].left_wall));

    CloseFile(f);                            {   }
end;

{------------------------------------------------------}

procedure ShowMaze(TheMaze : Maze);     {   }
var x, y          : Integer;
    Height, Width : Integer;            {     }
begin
    Width := High(TheMaze);             {     }
    Height := High(TheMaze[0]);

    with Form1.BackBuffer.Canvas do
    begin                                  {   }
        FillRect(Rect(0, 0, Form1.BackBuffer.Width, Form1.BackBuffer.Height));
        for x := 0 to Width - 1 do
            for y := 0 to Height - 1 do
            begin
                {       }
                if TheMaze[x, y].up_wall then
                begin
                    MoveTo(x * CellSize, y * CellSize);        {   }
                    LineTo((x + 1) * CellSize, y * CellSize);
                end;

                {       }
                if TheMaze[x, y].left_wall then
                begin
                    MoveTo(x * CellSize, y * CellSize);        {    }
                    LineTo(x * CellSize, (y + 1) * CellSize);
                end;
            end;
        MoveTo(0, Height * CellSize);                 {     }
        LineTo(Width * CellSize, Height * CellSize);  {    }
        LineTo(Width * CellSize, 0);
    end;
    {      }
    Form1.Screen.Canvas.CopyRect(Rect(0, 0, Form1.Screen.Width,
        Form1.Screen.Height), Form1.BackBuffer.Canvas,
        Rect(0, 0, Form1.Screen.Width, Form1.Screen.Height));
end;

{------------------------------------------------------}

{    }
procedure RecursiveSolve(TheMaze : Maze; xs, ys, xf, yf : Integer);
var Visited       : array of array of Boolean;  {    }
    x, y, xc, yc  : Integer;
    i             : Integer;
    Path          : array of TPoint;            {   }
    Height, Width : Integer;
const dx : array[1..4] of Integer = (1, 0, -1, 0);  {  }
      dy : array[1..4] of Integer = (0, -1, 0, 1);

{  : ,     
 (x, y)   (x + dx, y + dy),        }
function CanGo(x, y, dx, dy : Integer) : Boolean;
begin
    if dx = -1 then CanGo := not TheMaze[x, y].left_wall
    else if dx = 1 then CanGo := not TheMaze[x + 1, y].left_wall
    else if dy = -1 then CanGo := not TheMaze[x, y].up_wall
    else CanGo := not TheMaze[x, y + 1].up_wall;
end;

{      (x, y) }
function Solve(x, y, depth : Integer) : Boolean;
var i : Integer;
begin
    Visited[x, y] := true;                {     }
    Path[depth] := Point(x, y);           {      }
    Path[depth + 1] := Point(-1, -1);     {     }

    if (x = xf) and (y = yf) then         {     }
    begin
        Solve := true;                    {   }
        Exit;
    end;

    for i := 1 to 4 do
        {   ,    }
        if CanGo(x, y, dx[i], dy[i]) and not Visited[x + dx[i], y + dy[i]] then
            if Solve(x + dx[i], y + dy[i], depth + 1) then
            begin
                Solve := true;                {    }
                Exit;                         {   }
            end;

    Visited[x, y] := false;          {     }
    Solve := false;                  {    }
end;

begin                                                {   }
    Width := High(TheMaze);
    Height := High(TheMaze[0]);
    SetLength(Path, Height * Width + 1);  {     }
    SetLength(Visited, Width, Height);    {      }

    for x := 0 to Width - 1 do
        for y := 0 to Height - 1 do
            Visited[x, y] := false;       {      }

    if Solve(xs, ys, 0) then              {   ,   }
    begin
        i := 0;
        while not ((Path[i].X = -1) and (Path[i].Y = -1)) do
        begin
            xc := CellSize * (2 * Path[i].X + 1) div 2;
            yc := CellSize * (2 * Path[i].Y + 1) div 2;
            Form1.Screen.Canvas.Ellipse(xc - 5, yc - 5, xc + 5, yc + 5);
            i := i + 1;
        end;
    end;
end;

{------------------------------------------------------}

{    }
procedure WaveTracingSolve(TheMaze : Maze; xs, ys, xf, yf : Integer);
var Mark          : array of array of Integer;  {   }
    x, y, xc, yc  : Integer;
    N, i          : Integer;
    Height, Width : Integer;
const dx : array[1..4] of Integer = (1, 0, -1, 0);  {  }
      dy : array[1..4] of Integer = (0, -1, 0, 1);

{  : ,     
 (x, y)   (x + dx, y + dy),        }
function CanGo(x, y, dx, dy : Integer) : Boolean;
begin
    if dx = -1 then CanGo := not TheMaze[x, y].left_wall
    else if dx = 1 then CanGo := not TheMaze[x + 1, y].left_wall
    else if dy = -1 then CanGo := not TheMaze[x, y].up_wall
    else CanGo := not TheMaze[x, y + 1].up_wall;
end;

function Solve : Boolean;                     {   }
    var i, N, x, y : Integer;
        NoSolution : Boolean;
begin
    N := 1;                                   {     1 }

    repeat
        NoSolution := true;         {  ,    }
        for x := 0 to Width - 1 do
            for y := 0 to Height - 1 do
                if Mark[x, y] = N then   {  ,   N }
                    for i := 1 to 4 do   {    }
                        if CanGo(x, y, dx[i], dy[i]) and
                           (Mark[x + dx[i], y + dy[i]] = 0) then
                        begin            {      }
                            NoSolution := false;  {     }
                            {     N + 1 }
                            Mark[x + dx[i], y + dy[i]] := N + 1;
                            if (x + dx[i] = xf) and (y + dy[i] = yf) then
                            begin
                                Solve := true;   {     }
                                Exit;            {   }
                            end;
                        end;
        N := N + 1;                    {     }
    until NoSolution;     { ,      }

    Solve := false;     { ,    }
end;

begin
    Width := High(TheMaze);
    Height := High(TheMaze[0]);
    SetLength(Mark, Width, Height);    {     }

    for x := 0 to Width - 1 do         {     }
        for y := 0 to Height - 1 do
            Mark[x, y] := 0;

    Mark[xs, ys] := 1;         {     }
    if Solve then              {   ,   }
    begin
        x := xf; y := yf;
        for N := Mark[xf, yf] downto 1 do
        begin
            {       }
            xc := CellSize * (2 * x + 1) div 2;
            yc := CellSize * (2 * y + 1) div 2;
            Form1.Screen.Canvas.Ellipse(xc - 5, yc - 5, xc + 5, yc + 5);

            for i := 1 to 4 do
                if CanGo(x, y, dx[i], dy[i]) and
                   (Mark[x + dx[i], y + dy[i]] = N - 1) then
                begin
                    x := x + dx[i];   {     }
                    y := y + dy[i];
                    Break;
                end;
        end;
    end;
end;

{------------------------------------------------------}

procedure TForm1.LoadSaveBtnClick(Sender: TObject);  {   }
var TheMaze : Maze;
begin
    LoadMaze(TheMaze, 'maze1.txt');            {   }
    ShowMaze(TheMaze);                         {   }
    SaveMaze(TheMaze, 'maze2.txt');            {     }
end;

procedure TForm1.GoRecBtnClick(Sender: TObject);
var TheMaze : Maze;
begin
    LoadMaze(TheMaze, 'maze1.txt');
    ShowMaze(TheMaze);
    RecursiveSolve(TheMaze, 0, 0, 4, 0);
end;

procedure TForm1.GoWaveTraceBtnClick(Sender: TObject);
var TheMaze : Maze;
begin
    LoadMaze(TheMaze, 'maze1.txt');
    ShowMaze(TheMaze);
    WaveTracingSolve(TheMaze, 0, 0, 4, 0);
end;

{------------------------------------------------------}

{   }
function PrimGenerateMaze(Width, Height : Integer) : Maze;
type  AttrType = (Inside, Outside, Border);      {  " " }
var   TheMaze    : Maze;                         {   }
      x, y, i    : Integer;
      xc, yc     : Integer;
      xloc, yloc : Integer;
      Attribute  : array of array of AttrType;  {   }
      IsEnd      : Boolean;
      counter    : Integer;
const dx : array[1..4] of Integer = (1, 0, -1, 0);  {  }
      dy : array[1..4] of Integer = (0, -1, 0, 1);

label ExitFor1, ExitFor2, ExitFor3;         {   }

procedure BreakWall(x, y, dx, dy : Integer);             {   }
begin                                                    {   }
    if dx = -1 then TheMaze[x, y].left_wall := false
    else if dx = 1 then TheMaze[x + 1, y].left_wall := false
    else if dy = -1 then TheMaze[x, y].up_wall := false
    else TheMaze[x, y + 1].up_wall := false;
end;

begin
    SetLength(Attribute, Width, Height);    {     }
    SetLength(TheMaze, Width + 1, Height + 1);   {    }

    for x := 0 to Width - 1 do            {    }
        for y := 0 to Height - 1 do       {  Outside }
            Attribute[x, y] := Outside;

    for y := 0 to Height do                     {    }
        for x := 0 to Width do                  {  }
        begin
            TheMaze[x, y].left_wall := true;
            TheMaze[x, y].up_wall := true;
        end;

    Randomize;
    x := Random(Width);                    {    }
    y := Random(Height);
    Attribute[x, y] := Inside;             {     Inside }

    for i := 1 to 4 do                     {     }
    begin                                  {  Border }
        xc := x + dx[i];
        yc := y + dy[i];
        if (xc >= 0) and (yc >= 0) and (xc < Width) and (yc < Height) then
            Attribute[xc, yc] := Border;
    end;

    repeat                                       {   }
        IsEnd := true;
        counter := 0;
        for x := 0 to Width - 1 do              {   }
            for y := 0 to Height - 1 do      {    Border }
                if Attribute[x, y] = Border then counter := counter + 1;

        counter := Random(counter) + 1;               {    }
        for x := 0 to Width - 1 do                    {   }
            for y := 0 to Height - 1 do
                if Attribute[x, y] = Border then
                begin
                    counter := counter - 1;
                    if counter = 0 then
                    begin
                        xloc := x;             { xloc, yloc -   }
                        yloc := y;
                        goto ExitFor1;         {    }
                    end;
                end;
ExitFor1:
        Attribute[xloc, yloc] := Inside;      {    Inside }

        counter := 0;
        for i := 1 to 4 do
        begin
            xc := xloc + dx[i];
            yc := yloc + dy[i];
            if (xc >= 0) and (yc >= 0) and (xc < Width) and (yc < Height) then
            begin        {      Inside }
                if Attribute[xc, yc] = Inside then counter := counter + 1;
                if Attribute[xc, yc] = Outside then  {    }
                    Attribute[xc, yc] := Border;     { Outside  Border }
            end;
        end;

        counter := Random(counter) + 1;   {   Inside- }
        for i := 1 to 4 do
        begin
            xc := xloc + dx[i];
            yc := yloc + dy[i];
            if (xc >= 0) and (yc >= 0) and (xc < Width) and (yc < Height)
               and (Attribute[xc, yc] = Inside) then
            begin
                counter := counter - 1;
                if counter = 0 then         {      }
                begin                       {   }
                    BreakWall(xloc, yloc, dx[i], dy[i]);
                    goto ExitFor2;
                end;
            end;
        end;

ExitFor2:
        for x := 0 to Width - 1 do                { ,   }
            for y := 0 to Height - 1 do           {     }
                if Attribute[x, y] = Border then  {  Border }
                begin
                    IsEnd := false;               {  ,  }
                    goto ExitFor3;                {   }
                end;
ExitFor3:
        ShowMaze(TheMaze);                  {    }
        Application.ProcessMessages;
    until IsEnd;
    PrimGenerateMaze := TheMaze;
end;

{------------------------------------------------------}

{   }
function KruskalGenerateMaze(Width, Height : Integer) : Maze;
type Wall = record               {  "" }
    x, y, dx, dy : Integer;
end;
var   TheMaze   : Maze;           {   }
      Walls     : array of Wall;  {   }
      Temp      : array of Real;  {      }
      i, j      : Integer;
      tempw     : Wall;
      tempr     : Real;
      CurWall   : Wall;
      locations : Integer;
      counter   : Integer;

procedure BreakWall(x, y, dx, dy : Integer);             {   }
begin                                                    {   }
    if dx = -1 then TheMaze[x, y].left_wall := false
    else if dx = 1 then TheMaze[x + 1, y].left_wall := false
    else if dy = -1 then TheMaze[x, y].up_wall := false
    else TheMaze[x, y + 1].up_wall := false;
end;

function IsConnected(xs, ys, xf, yf : Integer) : Boolean;
var Mark          : array of array of Integer;  {   }
    x, y          : Integer;
    Height, Width : Integer;
const dx : array[1..4] of Integer = (1, 0, -1, 0);  {  }
      dy : array[1..4] of Integer = (0, -1, 0, 1);

{  : ,     
 (x, y)   (x + dx, y + dy),        }
function CanGo(x, y, dx, dy : Integer) : Boolean;
begin
    if dx = -1 then CanGo := not TheMaze[x, y].left_wall
    else if dx = 1 then CanGo := not TheMaze[x + 1, y].left_wall
    else if dy = -1 then CanGo := not TheMaze[x, y].up_wall
    else CanGo := not TheMaze[x, y + 1].up_wall;
end;

function Solve : Boolean;                     {   }
    var i, N, x, y : Integer;
        NoSolution : Boolean;
begin
    N := 1;                                   {     1 }

    repeat
        NoSolution := true;         {  ,    }
        for x := 0 to Width - 1 do
            for y := 0 to Height - 1 do
                if Mark[x, y] = N then   {  ,   N }
                    for i := 1 to 4 do   {    }
                        if CanGo(x, y, dx[i], dy[i]) and
                           (Mark[x + dx[i], y + dy[i]] = 0) then
                        begin            {      }
                            NoSolution := false;  {     }
                            {     N + 1 }
                            Mark[x + dx[i], y + dy[i]] := N + 1;
                            if (x + dx[i] = xf) and (y + dy[i] = yf) then
                            begin
                                Solve := true;   {     }
                                Exit;            {   }
                            end;
                        end;
        N := N + 1;                    {     }
    until NoSolution;     { ,      }

    Solve := false;     { ,    }
end;

begin
    Width := High(TheMaze);
    Height := High(TheMaze[0]);
    SetLength(Mark, Width, Height);    {     }

    for x := 0 to Width - 1 do         {     }
        for y := 0 to Height - 1 do
            Mark[x, y] := 0;

    Mark[xs, ys] := 1;         {     }
    IsConnected := Solve;
end;

begin
    {      }
    {   Width * Height 
    { (Width - 1) * Height + (Height - 1) * Width  }
    SetLength(Walls, (Width - 1) * Height + (Height - 1) * Width);
    SetLength(Temp, (Width - 1) * Height + (Height - 1) * Width);
    SetLength(TheMaze, Width + 1, Height + 1);   {    }

    for i := 0 to Width do                     {    }
        for j := 0 to Height do                  {  }
        begin
            TheMaze[i, j].left_wall := true;
            TheMaze[i, j].up_wall := true;
        end;

    Randomize;
    for i := 0 to (Width - 1) * Height + (Height - 1) * Width - 1 do
        Temp[i] := Random;      {   Temp   }

    counter := 0;                       {    }
    for i := 1 to Width - 1 do
        for j := 0 to Height - 1 do
        begin                                 {    }
            Walls[counter].x := i; Walls[counter].y := j;
            Walls[counter].dx := -1; Walls[counter].dy := 0;
            counter := counter + 1;
        end;
    for i := 0 to Width - 1 do
        for j := 1 to Height - 1 do
        begin                                 {    }
            Walls[counter].x := i; Walls[counter].y := j;
            Walls[counter].dx := 0; Walls[counter].dy := -1;
            counter := counter + 1;
        end;

    for i := 0 to (Width - 1) * Height + (Height - 1) * Width - 1 do
        for j := i to (Width - 1) * Height + (Height - 1) * Width - 1 do
            if Temp[i] > Temp[j] then           {    }
            begin
                tempr := Temp[i]; Temp[i] := Temp[j]; Temp[j] := tempr;
                tempw := Walls[i]; Walls[i] := Walls[j]; Walls[j] := tempw;
            end;

    locations := Width * Height;
    i := 0;
    while locations > 1 do               {   }
    begin                                {   }
        CurWall := Walls[i];
        i := i + 1;
        if not IsConnected(CurWall.x, CurWall.y,
                       CurWall.x + CurWall.dx, CurWall.y + CurWall.dy) then
        begin
            BreakWall(CurWall.x, CurWall.y, CurWall.dx, CurWall.dy);
            locations := locations - 1;
            ShowMaze(TheMaze);
            Application.ProcessMessages;
        end;
    end;

    KruskalGenerateMaze := TheMaze;
end;

{------------------------------------------------------}

procedure TForm1.PrimBtnClick(Sender: TObject);
var TheMaze : Maze;
begin
    TheMaze := PrimGenerateMaze(30, 18);
    SaveMaze(TheMaze, 'maze2.txt');
end;

procedure TForm1.KruskalBtnClick(Sender: TObject);
var TheMaze : Maze;
begin
    TheMaze := KruskalGenerateMaze(30, 18);
    SaveMaze(TheMaze, 'maze2.txt');
end;

end.
